# Generated by Django 4.2 on 2024-04-25 07:33

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):
    dependencies = [
        ("tests", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="NamedCollection",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=25, unique=True)),
                ("code", models.UUIDField(default=uuid.uuid4, unique=True)),
            ],
        ),
    ]
